import gradio as gr

from groovy.flow import Flow


def test_flow_with_simple_task():
    """Test Flow with a basic task string"""
    flow = Flow(task="Hello World", agent_fn=lambda x: x)
    assert flow.task == "Hello World"
    assert len(flow.inputs) == 0
    assert flow.agent_fn is not None


def test_flow_with_gradio_inputs():
    """Test Flow initialization with Gradio input components"""
    text_input = gr.Textbox(label="Input")
    slider = gr.Slider(minimum=0, maximum=100)
    flow = Flow(
        task="Process {text} with value {value}",
        inputs=[text_input, slider],
        agent_fn=lambda x: x,
    )

    assert len(flow.inputs) == 2
    assert isinstance(flow.inputs[0], gr.Textbox)
    assert isinstance(flow.inputs[1], gr.Slider)


def test_flow_serialization():
    """Test Flow serialization to and from JSON"""
    text_input = gr.Textbox(label="Test Input")
    original_flow = Flow(
        task="Process {text}", inputs=[text_input], agent_fn=lambda x: x
    )

    json_data = original_flow.to_json()
    restored_flow = Flow.from_json(json_data, agent_fn=lambda x: x)

    assert restored_flow.task == original_flow.task
    assert len(restored_flow.inputs) == len(original_flow.inputs)
    assert isinstance(restored_flow.inputs[0], gr.Textbox)
    assert restored_flow.inputs[0].label == "Test Input"



from pathlib import Path

import gradio as gr
from gradio.themes.utils import colors

from groovy.app_utils import add_step_counter
from groovy.types import Image, StepMessage, raw_value_to_step_message

theme = gr.themes.Base(
    primary_hue="purple",
)

theme.set(
    button_cancel_background_fill=colors.red.c500,
    button_cancel_background_fill_dark=colors.red.c700,
    button_cancel_background_fill_hover=colors.red.c600,
    button_cancel_background_fill_hover_dark=colors.red.c800,
)


def create_app(self, inputs, prompt, streamer, run_immediately, save_recording):
    with gr.Blocks(theme=theme) as app:
        with gr.Accordion(
            "Input parameters (used to construct the task)",
            open=True,
            visible=bool(inputs),
        ) as inputs_accordion:
            for input in inputs:
                input.render()

        with gr.Group():
            prompt_box = gr.Textbox(label="🕺 Constructed Task", value=prompt)
            with gr.Row():
                run_button = gr.Button("Start Flow", variant="primary")
                stop_button = gr.Button("Stop Flow", variant="stop", visible=False)

        chat_log = gr.Chatbot(
            label="Log",
            type="messages",
            group_consecutive_messages=False,
            visible=False,
        )

        @gr.on(
            triggers=[app.load] + [input.change for input in inputs],
            inputs=inputs,
            outputs=[prompt_box],
            trigger_mode="always_last",
        )
        def construct_prompt(*input_values):
            return prompt.format(*input_values)

        run_triggers = [run_button.click]
        if run_immediately:
            run_triggers.append(app.load)

        def run_flow_ui_changes():
            return {
                run_button: gr.Button(value="Running...", interactive=False),
                chat_log: gr.Chatbot(visible=True, type="messages"),
                inputs_accordion: gr.Accordion(open=False),
            }

        def reset_ui():
            return {
                run_button: gr.Button(value="Start Flow", interactive=True),
                inputs_accordion: gr.Accordion(open=True, visible=bool(inputs)),
            }

        def run_flow(prompt):
            chat_messages = [gr.ChatMessage(content=prompt, role="user")]
            yield chat_messages

            images_for_gif = []

            for step_message in streamer(prompt):
                if not isinstance(step_message, StepMessage):
                    step_message = raw_value_to_step_message(step_message)
                chat_messages.append(step_message.as_chat_message())

                # Any returned gv.Image messages are automatically added to the recording
                if (
                    isinstance(step_message, Image)
                    and step_message.save_in_recording
                    and save_recording
                ):
                    img_with_text = add_step_counter(
                        step_message.value, len(images_for_gif) + 1
                    )
                    images_for_gif.append(img_with_text)
                    if len(images_for_gif) > 0:
                        gif_dir = Path.cwd() / ".groovy"
                        gif_dir.mkdir(exist_ok=True)
                        gif_path = gif_dir / "recording.gif"
                        images_for_gif[0].save(
                            gif_path,
                            save_all=True,
                            append_images=images_for_gif[1:],
                            duration=500,
                            loop=0,
                        )
                yield chat_messages

        gr.on(
            fn=run_flow_ui_changes,
            triggers=run_triggers,
            outputs=[run_button, stop_button, chat_log, inputs_accordion],
        ).then(
            fn=run_flow,
            inputs=[prompt_box],
            outputs=[chat_log],
        ).then(
            fn=reset_ui,
            outputs=[run_button, stop_button, chat_log, inputs_accordion],
        )

    return app



import gradio as gr

import groovy as gv

flow = gv.Flow(
    # task="Find me some events in San Francisco related to board games using Meetup.",
    task="Find me some events in {} related to {} using Meetup.",
    inputs=[
        gr.Dropdown(
            ["San Francisco", "New York", "Chicago", "Los Angeles", "Boston"],
            allow_custom_value=True,
        ),
        gr.Dropdown(
            ["board games", "cooking", "hiking", "reading", "writing"],
            allow_custom_value=True,
        ),
    ],
)

if __name__ == "__main__":
    flow.launch()



import time
from collections.abc import Generator
from typing import Callable, Sequence, Union

import gradio as gr
from gradio_client import Client

from groovy.types import Image, String
from groovy.utils import resolve_space_url_to_id


class Flow:
    """
    The core class in Groovy. A Flow requires three things:
    * A task to run (string or format string)
    * A set of input components (optional)
    * A generator function that accepts a task string and yields responses.

    After instantiating a Flow, you can call .launch() to launch the flow in a Gradio app.
    """

    def __init__(
        self,
        task: str,
        inputs: Sequence[gr.Component] | None = None,
        agent_fn: Callable[[str], Generator[Union[String, Image], None, None]]
        | None = None,
    ):
        """
        Parameters:
            task: The task to run. Can be a regular string or a format string, in which case the input components' values will be passed to it.
            inputs: The input components whose values will be passed to the task, if it's a format string.
            agent_fn: Fancy name for a generator function that accepts a task string and yields an arbitrary number of `gv.String()` or `gv.Image()` responses. If not provided, Groovy includes a built-in browser agent function, which browses the web to complete a task.
        """
        # Import here to speed up the import time of the groovy module

        self.task = task
        self.inputs = inputs or []
        if agent_fn is None:
            try:
                from groovy.agent import browser_agent_fn
            except ImportError:
                raise ImportError(
                    "The browser agent function is not available. Please install the full version of Groovy: `pip install groovy[full]`."
                )
            self.agent_fn = browser_agent_fn
        else:
            self.agent_fn = agent_fn

    def launch(
        self,
        save_recording: bool = True,
        run_immediately: bool = False,
        **kwargs,
    ):
        """
        Launches the flow in a Gradio app.

        Parameters:
            save_recording: If True, collects any images generated by the flow and saves them as a gif in the working directory.
            run_immediately: Whether to run the task immediately at .launch() or whether to wait for the user to click the "Run" button in the Gradio app.
            kwargs: Additional keyword arguments to pass to the Gradio app's launch() method.
        """
        # Import here to speed up the import time of the groovy module
        from groovy.app import create_app
        from groovy.utils import open_positioned_browser

        self.save_recording = save_recording
        self.run_immediately = run_immediately
        self.app = create_app(
            self,
            self.inputs,
            self.task,
            self.agent_fn,
            self.run_immediately,
            self.save_recording,
        )
        _, self.url, _ = self.app.launch(
            inline=False, inbrowser=False, prevent_thread_lock=True
        )
        open_positioned_browser(self.url, width=500)
        while 1:
            time.sleep(0.001)
        return self.url

    def run(self, *args):
        output = None
        for step in self.agent_fn(self.task, *args):
            output = step
        return output

    def to_json(self) -> dict:
        data = {
            "task": self.task,
            "inputs": [input_comp.get_config() for input_comp in self.inputs],
        }
        return data

    @staticmethod
    def component_from_config(config: dict) -> gr.Component:
        cls = gr.utils.component_or_layout_class(config.pop("name"))
        constructor_args = cls.recover_kwargs(config)
        return cls(**constructor_args)

    @classmethod
    def from_json(
        cls,
        data: dict,
        agent_fn: Callable[[str], Generator[Union[String, Image], None, None]]
        | None = None,
    ):
        return cls(
            task=data["task"],
            inputs=[
                cls.component_from_config(input_config)
                for input_config in data["inputs"]
            ],
            agent_fn=agent_fn,
        )

    @classmethod
    def from_space(
        cls,
        space_url: str,
        agent_fn: Callable[[str], Generator[Union[String, Image], None, None]]
        | None = None,
    ):
        space_id = resolve_space_url_to_id(space_url)
        client = Client(space_id)
        config = client.predict(api_name="/flow_config")
        return cls.from_json(config, agent_fn)



import warnings

import gradio as gr
import PIL.Image


class StepMessage:
    """
    An abstract class that represents a message yielded by a step in a flow.

    This is for two purposes:
    1. To provide a common interface for values that should be yielded by a step in a flow
    2. To easily convert values to a gradio ChatMessage object
    """

    def __init__(self, value):
        self.value = value

    def as_chat_message(self) -> gr.ChatMessage:
        raise NotImplementedError("Subclass must implement as_chat_message")


class String(StepMessage):
    """
    A step message that behaves like a string.
    """

    def as_chat_message(self) -> gr.ChatMessage:
        return gr.ChatMessage(content=self.value, role="assistant")


class Image(StepMessage):
    """
    A step message that contains an PIL Image.
    """

    def __init__(self, value, save_in_recording: bool = True):
        self.value = value
        self.save_in_recording = save_in_recording

    def as_chat_message(self) -> gr.ChatMessage:
        return gr.ChatMessage(content=gr.Image(self.value), role="assistant")


def raw_value_to_step_message(value):
    if isinstance(value, str):
        return String(value)
    elif isinstance(value, PIL.Image.Image):
        return Image(value)
    else:
        warnings.warn(f"Unsupported value type: {type(value)}. Will treat as a string.")
        return String(str(value))



import gradio as gr
import groovy as gv
from flow import flow

with gr.Blocks() as app:
    task_box = gr.Textbox(label="🕺 Task", value="flow.task", info="Run this workflow locally by installing the [Groovy Python package](https://github.com/abidlabs/groovy) and then running `groovy run https://huggingface.co/space/url>`.")
    with gr.Row():
        if flow.inputs:
            with gr.Column(scale=1):
                for component in flow.inputs:
                    component.render()
        with gr.Column(scale=2):
            gr.Image(label="Recording", value=".groovy/recording.gif")
            config = gr.JSON(visible=False)

    @gr.on(
        triggers=[app.load] + [input.change for input in flow.inputs],
        inputs=flow.inputs,
        outputs=[task_box],
        trigger_mode="always_last",
        show_api=False
    )
    def construct_prompt(*input_values):
        return flow.task.format(*input_values)

    app.load(flow.to_json, None, config, api_name="flow_config")


if __name__ == "__main__":
    app.launch()



import os
from io import BytesIO
from time import sleep

import helium
from dotenv import load_dotenv
from PIL import Image
from screeninfo import get_monitors
from selenium import webdriver
from selenium.common.exceptions import ElementNotInteractableException, TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from smolagents import CodeAgent, LiteLLMModel, tool
from smolagents.agents import ActionStep

load_dotenv()

api_key = os.getenv("OPENAI_API_KEY")

agent = None
model = LiteLLMModel(
    model_id="gpt-4o",
    api_key=api_key,
)


def take_screenshot(current_step: ActionStep, agent: CodeAgent) -> Image.Image:
    """Captures a screenshot of the current browser window,

    Returns:
        Image.Image: PIL Image object containing the screenshot
    """
    sleep(0.5)
    driver = agent.driver

    for step in agent.logs:  # Remove previous screenshots from logs for lean processing
        if (
            isinstance(step, ActionStep)
            and step.step_number <= current_step.step_number - 2
        ):
            step.observations_images = None

    png_bytes = driver.get_screenshot_as_png()
    image = Image.open(BytesIO(png_bytes))
    current_step.observations_images = [
        image.copy()
    ]  # Create a copy to ensure it persists, important!

    url_info = f"Current url: {driver.current_url}"
    step.observations = (
        url_info if step.observations is None else step.observations + "\n" + url_info
    )
    return image


@tool
def search_item_ctrl_f(text: str, nth_result: int = 1) -> str:
    """
    Searches for text on the current page via Ctrl + F and jumps to the nth occurrence.
    Args:
        text: The text to search for
        nth_result: Which occurrence to jump to (default: 1)
    """
    driver = agent.driver
    elements = driver.find_elements(By.XPATH, f"//*[contains(text(), '{text}')]")
    if nth_result > len(elements):
        raise Exception(
            f"Match n°{nth_result} not found (only {len(elements)} matches found)"
        )
    result = f"Found {len(elements)} matches for '{text}'."
    elem = elements[nth_result - 1]
    driver.execute_script("arguments[0].scrollIntoView(true);", elem)
    result += f"Focused on element {nth_result} of {len(elements)}"
    return result


@tool
def go_back() -> None:
    """Goes back to previous page."""
    driver = agent.driver
    driver.back()


@tool
def close_popups() -> str:
    """
    Closes any visible modal or pop-up on the page. Use this to dismiss pop-up windows! This does not work on cookie consent banners.
    """
    # Common selectors for modal close buttons and overlay elements
    modal_selectors = [
        "button[class*='close']",
        "[class*='modal']",
        "[class*='modal'] button",
        "[class*='CloseButton']",
        "[aria-label*='close']",
        ".modal-close",
        ".close-modal",
        ".modal .close",
        ".modal-backdrop",
        ".modal-overlay",
        "[class*='overlay']",
    ]
    driver = agent.driver
    wait = WebDriverWait(driver, timeout=0.5)

    for selector in modal_selectors:
        try:
            elements = wait.until(
                EC.presence_of_all_elements_located((By.CSS_SELECTOR, selector))
            )

            for element in elements:
                if element.is_displayed():
                    try:
                        driver.execute_script("arguments[0].click();", element)
                    except ElementNotInteractableException:
                        element.click()

        except TimeoutException:
            continue
        except Exception as e:
            print(f"Error handling selector {selector}: {str(e)}")
            continue
    return "Modals closed"


def create_agent() -> CodeAgent:
    """Creates and returns a configured CodeAgent with initialized Chrome driver."""
    primary_monitor = get_monitors()[0]
    screen_height = primary_monitor.height
    screen_width = primary_monitor.width
    chrome_options = webdriver.ChromeOptions()
    chrome_options.add_argument("--force-device-scale-factor=1")
    chrome_options.add_argument(f"--window-size={screen_width - 550},{screen_height}")
    chrome_options.add_argument("--disable-pdf-viewer")
    chrome_options.add_argument("--window-position=550,0")

    driver = helium.start_chrome(headless=False, options=chrome_options)

    agent = CodeAgent(
        tools=[go_back, close_popups, search_item_ctrl_f],
        model=model,
        additional_authorized_imports=["helium"],
        max_steps=50,
        verbosity_level=0,
    )
    agent.driver = driver  # Store the driver instance in the agent
    return agent


helium_instructions = """
You can use helium to access websites. Don't bother about the helium driver, it's already managed.
First you need to import everything from helium, then you can do other actions!
Code:
```py
from helium import *
go_to('github.com/trending')
```<end_code>

You can directly click clickable elements by inputting the text that appears on them.
Code:
```py
click("Top products")
```<end_code>

If it's a link:
Code:
```py
click(Link("Top products"))
```<end_code>

If you try to interact with an element and it's not found, you'll get a LookupError.
In general stop your action after each button click to see what happens on your screenshot.
Never try to login in a page.

To scroll up or down, use scroll_down or scroll_up with as an argument the number of pixels to scroll from.
Code:
```py
scroll_down(num_pixels=1200) # This will scroll one viewport down
```<end_code>

When you have pop-ups with a cross icon to close, don't try to click the close icon by finding its element or targeting an 'X' element (this most often fails).
Just use your built-in tool `close_popups` to close them:
Code:
```py
close_popups()
```<end_code>

You can use .exists() to check for the existence of an element. For example:
Code:
```py
if Text('Accept cookies?').exists():
    click('I accept')
```<end_code>

Proceed in several steps rather than trying to solve the task in one shot.
And at the end, only when you have your answer, return your final answer.
Code:
```py
final_answer("YOUR_ANSWER_HERE")
```<end_code>

If pages seem stuck on loading, you might have to wait, for instance `import time` and run `time.sleep(2.0)`. But don't overuse this!
To list elements on page, DO NOT try code-based element searches like 'contributors = find_all(S("ol > li"))': just look at the latest screenshot you have and read it visually, or use your tool search_item_ctrl_f.
Of course, you can act on buttons like a user would do when navigating.
After each code blob you write, you will be automatically provided with an updated screenshot of the browser and the current browser url.
But beware that the screenshot will only be taken at the end of the whole action, it won't see intermediate states.
Don't kill the browser.
"""

search_request = """
Find flights from New York to San Francisco on 2025-02-01. Give me the cheapest flight.
"""


def browser_agent_fn(prompt: str):
    global api_key
    if not api_key:
        api_key = input("Please enter your OpenAI API key: ").strip()
        if not api_key:
            raise ValueError("OpenAI API key is required")
        model.api_key = api_key

    global agent
    if agent is None:
        agent = create_agent()
    for step in agent.run(prompt + helium_instructions, stream=True):
        if isinstance(step, ActionStep):
            yield str(step.llm_output.strip())
            yield take_screenshot(step, agent)
        else:
            yield "**Final Answer:**\n\n" + str(step)  # Return the final answer
            return


if __name__ == "__main__":
    browser_agent_fn(search_request)



from PIL import ImageDraw, ImageFont
from PIL.Image import Image

from groovy.utils import open_positioned_browser


def add_step_counter(image: Image, step_number: int) -> Image:
    """Add a step counter to the bottom left of an image.

    Args:
        image: The PIL Image to modify
        step_number: The step number to display

    Returns:
        A new Image with the step counter added
    """
    img_with_text = image.copy()
    draw = ImageDraw.Draw(img_with_text)

    font = ImageFont.load_default().font_variant(size=24)

    step_text = f"Step {step_number}"
    text_bbox = draw.textbbox((0, 0), step_text, font=font)
    text_width = text_bbox[2] - text_bbox[0]
    text_height = text_bbox[3] - text_bbox[1]

    padding = 8
    x = padding
    y = img_with_text.height - text_height - padding - 2

    draw.rectangle(
        [
            0,
            img_with_text.height - text_height - padding * 2,
            x + text_width + padding * 2,
            img_with_text.height,
        ],
        fill="#d8b4fe",  # Purple color
        outline="#d8b4fe",
    )

    draw.text((x, y), step_text, fill="black", font=font)

    return img_with_text


if __name__ == "__main__":
    open_positioned_browser("https://www.google.com")
    import time

    time.sleep(3)



from urllib.parse import urlparse

from screeninfo import get_monitors
from selenium import webdriver
from selenium.webdriver.chrome.options import Options


def open_positioned_browser(url, width=200, x=0, y=0):
    """
    Opens Chrome browser with specified width and full height of the primary monitor using Selenium.

    Args:
        url (str): The URL to open
        width (int): Window width in pixels
        x (int): X position on screen
        y (int): Y position on screen
    """
    # Get primary monitor dimensions
    primary_monitor = get_monitors()[0]
    screen_height = primary_monitor.height

    chrome_options = Options()
    chrome_options.add_argument(f"--window-size={width},{screen_height}")
    chrome_options.add_argument(f"--window-position={x},{y}")
    chrome_options.add_experimental_option("detach", True)

    driver = webdriver.Chrome(options=chrome_options)
    driver.get(url)
    return driver


if __name__ == "__main__":
    open_positioned_browser("https://www.google.com")
    import time

    time.sleep(3)


def resolve_space_url_to_id(space_url: str) -> str:
    """
    Resolves a Hugging Face space URL to its space ID.

    Parameters:
        space_url (str): The URL of the Hugging Face space
            e.g., 'https://huggingface.co/spaces/abidlabs/examples'
    Returns:
        str: The space ID in the format 'owner/space_name'
            e.g., 'abidlabs/examples'
    Raises:
        ValueError: If the URL is not a valid Hugging Face space URL
        ConnectionError: If there are issues connecting to the Hugging Face API
    """
    parsed_url = urlparse(space_url)
    if parsed_url.netloc not in ["huggingface.co", "www.huggingface.co"]:
        raise ValueError("Not a valid Hugging Face Space URL")
    path_parts = parsed_url.path.strip("/").split("/")
    if len(path_parts) < 3 or path_parts[0] != "spaces":
        raise ValueError("Not a valid Hugging Face Space URL")
    owner = path_parts[1]
    space_name = path_parts[2]
    return f"{owner}/{space_name}"



from pathlib import Path

from groovy.flow import Flow
from groovy.types import Image, String

__version__ = Path(__file__).parent.joinpath("version.txt").read_text().strip()

__all__ = ["Flow", "__version__", "Image", "String"]
